#!/usr/bin/env python3

import cgi, cgitb
import re, json, yaml
from os import environ, path, pardir
import sys, datetime, argparse
from pymongo import MongoClient
from operator import itemgetter

pkg_path = path.join( path.dirname( path.abspath(__file__) ), pardir, pardir )
sys.path.append( pkg_path )
from bycon import *

"""podmd

podmd"""

################################################################################
################################################################################
################################################################################

def main():

    publications()
    
################################################################################

def publications():

    initialize_bycon_service(byc)

    get_filter_flags(byc)
    parse_filters(byc)

    create_empty_service_response(byc)

    # data retrieval & response population
    query, e = _create_filters_query( byc )
    response_add_error(byc, 422, e )

    geo_q, geo_pars = geo_query( byc )

    if geo_q:
        for g_k, g_v in geo_pars.items():
            response_add_received_request_summary_parameter(byc, g_k, g_v)
        if len(query.keys()) < 1:
            query = geo_q
        else:
            query = { '$and': [ geo_q, query ] }

    if len(query.keys()) < 1:
        response_add_error(byc, 422, "No query could be constructed from the parameters provided." )

    cgi_break_on_errors(byc)

    mongo_client = MongoClient( )
    pub_db = byc["config"]["info_db"]
    pub_coll = mongo_client[ pub_db ][ "publications" ]

    p_re = re.compile( byc["filter_definitions"]["PMID"]["pattern"] )

    p_l = [ ]
    d_k = collations_set_delivery_keys(byc)
    
    for pub in pub_coll.find( query, { "_id": 0 } ):
        s = { }
        if len(d_k) < 1:
            s = pub
        else:
            for k in d_k:
                # TODO: harmless hack
                if k in pub.keys():
                    if k == "counts":
                        s[ k ] = { }
                        for c in pub[ k ]:
                            if pub[ k ][ c ]:
                                try:
                                    s[ k ][ c ] = int(float(pub[ k ][ c ]))
                                except:
                                    s[ k ][ c ] = 0
                            else:
                                s[ k ][ c ] = 0
                    else:
                        s[ k ] = pub[ k ]
                else:
                    s[ k ] = None
        try:
            s_v = p_re.match(s[ "id" ]).group(3)
            s[ "sortid" ] = int(s_v)
        except:
            s[ "sortid" ] = -1

        p_l.append( s )

    mongo_client.close( )
 
    results = sorted(p_l, key=itemgetter('sortid'), reverse = True)

    check_map_response(byc, results)

    populate_service_response( byc, results)
    cgi_print_response( byc, 200 )

################################################################################
################################################################################

def check_map_response(byc, results):

    if not "map" in byc["output"]:
        return

    u_locs = {}

    for p in results:
        if not "provenance" in p:
            pass
        if not "counts" in p:
            pass

        geoloc = p["provenance"].get("geo_location", None)
        if geoloc is None:
            pass

        l_k = "{}::{}".format(geoloc["geometry"]["coordinates"][1], geoloc["geometry"]["coordinates"][0])

        if not l_k in u_locs.keys():
            u_locs.update({l_k:{"geo_location": geoloc}})
            u_locs[l_k]["geo_location"]["properties"].update({"items":[]})

        m_c = p["counts"].get("genomes", 0)
        m_s = u_locs[l_k]["geo_location"]["properties"].get("marker_count", 0) + m_c
        # print(m_c, m_s)

        i = "<a href='/publication/?id={}'>{}</a> ({})".format(p["id"], p["id"], m_c)
        u_locs[l_k]["geo_location"]["properties"].update({"marker_count":m_s})
        u_locs[l_k]["geo_location"]["properties"]["items"].append(i)

    geolocs =  u_locs.values()

    print_map_from_geolocations(byc, geolocs)

################################################################################

def _create_filters_query( byc ):

    query = { }
    error = ""

    q_list = [ ]
    count_pat = re.compile( r'^(\w+?)\:([>=<])(\d+?)$' )

    # TODO: This doesn't apply any more?

    for f in byc[ "filters" ]:
        f_val = f["id"]
        pre_code = re.split('-|:', f_val)
        pre = pre_code[0]
        dbk = byc[ "filter_definitions" ][ pre ][ "db_key" ]

        if count_pat.match( f_val ):
            pre, op, no = count_pat.match(f_val).group(1,2,3)
            dbk = byc[ "filter_definitions" ][ pre ][ "db_key" ]
            if op == ">":
                op = '$gt'
            elif op == "<":
                op = '$lt'
            elif op == "=":
                op = '$eq'
            else:
                error = "uncaught filter error: {}".format(f_val)
                continue
            q_list.append( { dbk: { op: int(no) } } )
        elif "start" in byc[ "filter_flags" ][ "precision" ] or len(pre_code) == 1:
            """podmd
            If there was only prefix a regex match is enforced - basically here
            for the selection of PMID labeled publications.
            podmd"""
            q_list.append( { "id": re.compile(r'^'+f_val ) } )
        elif pre in byc[ "filter_definitions" ].keys():
            # TODO: hacky method for pgxuse => redo
            q_v = f_val
            try:
                if byc[ "filter_definitions" ][ pre ][ "remove_prefix" ] is True:
                    q_v = pre_code[1]
            except:
                pass
            q_list.append( { dbk: q_v } )
        else:
            q_list.append( { "id": f_val } )

    query = create_and_or_query_for_list('$and', q_list)

    return query, error

################################################################################
################################################################################

if __name__ == '__main__':
    main()



#!/usr/bin/env python3

import cgi, cgitb
import re, json, yaml
from os import environ, pardir, path, scandir
import sys, datetime
from humps import camelize

pkg_path = path.join( path.dirname( path.abspath(__file__) ), pardir, pardir )
sys.path.append( pkg_path )
from bycon import *

"""podmd

* <https://progenetix.org/services/schemas/biosample>

podmd"""

################################################################################
################################################################################
################################################################################

def main():

    endpoints()
    
################################################################################

def endpoints():

    initialize_bycon_service(byc)
    # create_empty_service_response(byc)

    schema_name = rest_path_value("endpoints")
    comps = schema_name.split('.')
    schema_name = comps.pop(0)

    # if "empty_value" in schema_name:
    #     schema_name = "biosample"

    if "empty_value" in schema_name:
        p = path.join( pkg_path, "schemas", "models", "json", "progenetix-model", "endpoints.json")
    else:
        p = path.join( pkg_path, "schemas", "models", "json", "progenetix-model", schema_name, "endpoints.json")

    root_def = RefDict(p)
    exclude_keys = [ "format", "examples" ]
    s = materialize(root_def, exclude_keys = exclude_keys)

    if not s is False:

        print('Content-Type: application/json')
        print('status:200')
        print()
        print(json.dumps(camelize(s), indent=4, sort_keys=True, default=str)+"\n")
        exit()
    
################################################################################
################################################################################

if __name__ == '__main__':
    main()



import cgi, cgitb
import re, yaml
from pymongo import MongoClient
from bson.son import SON

from cgi_parsing import *
from filter_parsing import *
from variant_parsing import *

################################################################################

def initialize_beacon_queries(byc):

    byc.update({"queries": { }})

    get_filter_flags(byc)
    parse_filters(byc)

    parse_variant_parameters(byc)
    get_variant_request_type(byc)

    # generate_queries(byc)
    select_dataset_ids(byc)

    if len(byc["dataset_ids"]) < 1:
        print_text_response("No existing dataset_id - please check dataset_definitions")

    return byc

################################################################################

def generate_queries(byc, ds_id="progenetix"):

    byc.update({"queries": { }})

    _update_queries_from_path_id( byc )
    _update_queries_from_id_values( byc )
    _update_queries_from_cohorts_query(byc)
    _update_queries_from_filters( byc, ds_id )
    _update_queries_from_variants( byc )
    _update_queries_from_geoquery( byc )
    _update_queries_from_hoid( byc)
    _replace_queries_in_test_mode( byc)
    _purge_empty_queries( byc )

    # TODO: HOT FIX
    if "runs" in byc["queries"].keys():
        if not "callsets" in byc["queries"].keys():
            byc["queries"]["callsets"] = byc["queries"].pop("runs")

    return byc

################################################################################

def _purge_empty_queries( byc ):

    empties = [ ]
    for k, v in byc["queries"].items():
        if not v:
            empties.append( k )
    for e_k in empties:
        byc["queries"].pop(k, None)

    return byc

################################################################################

def _replace_queries_in_test_mode(byc):

    if byc["test_mode"] is not True:
        return byc

    try:
        collname = byc["response_entity"]["collection"]
    except:
        return byc

    ret_no = int(byc.get('test_mode_count', 5))

    ds_id = byc["dataset_ids"][0]
    mongo_client = MongoClient()
    data_db = mongo_client[ ds_id ]
    data_collnames = data_db.list_collection_names()

    if not collname in data_collnames:
        return byc

    data_coll = mongo_client[ ds_id ][ collname ]
    rs = list( data_coll.aggregate([{"$sample": {"size":ret_no}}]) )

    _ids = []
    for r in rs:
        _ids.append(r["_id"])

    byc.update( {
        "queries": { collname: { "_id": {"$in": _ids } } },
        "empty_query_all_count": data_coll.estimated_document_count()
    } )

    return byc

################################################################################

def _update_queries_from_path_id( byc ):

    b_mps = byc["beacon_mappings"]

    if "service" in byc["request_path_root"]:
        b_mps = byc["services_mappings"]

    if not byc["request_entity_id"]:
        return byc

    r_e_id = byc["request_entity_id"]
    p_id_v = byc["request_entity_path_id_value"]

    if not byc["request_entity_path_id_value"]:
        return byc

    if not r_e_id in b_mps["response_types"]:
        return byc

    collname = b_mps["response_types"][r_e_id]["collection"]

    if not collname:
        return byc

    byc["queries"].update( { collname: { "id": p_id_v } } )

    return byc

################################################################################

def _update_queries_from_cohorts_query(byc):

    if not "cohorts" in byc["queries"]:
        return byc

    if "cohort" in byc["response_entity_id"]:
        return byc

    c_q = byc["queries"]["cohorts"]

    query = {}
    if "id" in c_q:
        query = {"cohorts.id": c_q["id"]}

    byc["queries"].pop("cohorts", None)

    update_query_for_scope( byc, query, "biosamples")

    return byc

################################################################################

def _update_queries_from_id_values(byc):

    id_f_v = byc["beacon_mappings"]["id_queryscope_mappings"]
    f_d = byc["form_data"]

    this_id_k = byc["response_entity_id"]+"_ids"

    if "ids" in f_d:
        if not this_id_k in f_d:
            f_d.update({this_id_k: f_d["ids"]})

    for id_k, id_s in id_f_v.items():
        q = False
        if id_k in f_d:
            id_v = f_d[id_k]
            if len(id_v) > 1:
                q = {"id":{"$in":id_v}}
            elif len(id_v) == 1:
                q = {"id":id_v[0]}
        if q is not False:
            update_query_for_scope( byc, q, id_s, "AND")

    return byc

################################################################################

def _update_queries_from_hoid( byc):

    if "accessid" in byc["form_data"]:

        accessid = byc["form_data"]["accessid"]
        ho_client = MongoClient()
        ho_db = ho_client[ byc["config"]["info_db"] ]
        ho_coll = ho_db[ byc["config"][ "handover_coll" ] ]
        h_o = ho_coll.find_one( { "id": accessid } )
        
        # accessid overrides ... ?
        if h_o:
            t_k = h_o["target_key"]
            t_v = h_o["target_values"]
            c_n = h_o["target_collection"]
            t_c = h_o["target_count"]

            byc.update({"original_queries": h_o.get("original_queries", None)})

            set_pagination_range(t_c, byc)
            t_v = paginate_list(t_v, byc)

            t_db = h_o["source_db"]
            h_o_q = { t_k: { '$in': t_v } }
            if c_n in byc["queries"]:
                byc["queries"].update( { c_n: { '$and': [ h_o_q, byc["queries"][ c_n ] ] } } )
            else:
                byc["queries"].update( { c_n: h_o_q } )

    return byc

################################################################################

def _update_queries_from_filters(byc, ds_id="progenetix"):

    """The new version assumes that dataset_id, scope (collection) and field are
    stored in the collation entries. Only filters with exact match to an entry
    in the lookup "collations" collection will be evaluated.
    While the Beacon v2 protocol assumes a logical `AND` between filters, bycon
    has a slightly differing approach:
    * filters against the same field (in the same collection) are treated as
    logical `OR` since this seems logical; and also allows the use of the same
    query object for hierarchical (`child_terms`) query expansion
    * the bycon API allows to pass a `filterLogic` parameter with either `AND`
    (default value) or `OR`

    CAVE:   Filters are assumed to be id values in thje collation collection
            OR have a "collationed" flag in the filter_definitions set to False
    """

    f_defs = byc["filter_definitions"]
    f_lists = {}

    logic = byc[ "filter_flags" ][ "logic" ]
    f_desc = byc[ "filter_flags" ][ "descendants" ]
    # precision = byc[ "filter_flags" ][ "precision" ]

    mongo_client = MongoClient()
    coll_coll = mongo_client[ ds_id ][ byc["config"]["collations_coll"] ]

    filters = byc.get("filters", [])

    for f in filters:

        f_val = f["id"]
        f_desc = f.get("includeDescendantTerms", f_desc)
        f_scope = f.get("scope", False)

        f_info = coll_coll.find_one({"id": f["id"]})

        if f_info is None:

            for f_d in f_defs.values():
                f_re = re.compile(f_d["pattern"])
                if f_re.match(f["id"]):
                    if f_d["collationed"] is False:
                        f_info = {
                            "id": f["id"],
                            "scope": f_d["scope"],
                            "db_key": f_d["db_key"],
                            "child_terms": [f["id"]]
                        }
                        f_desc = False

        if f_info is None:
            continue

        f_field = f_info.get("db_key", "id")

        if f_scope is False:
            f_scope = f_info["scope"]

        if f_scope not in byc[ "config" ][ "queried_collections" ]:
            continue

        if f_scope not in f_lists.keys():
            f_lists.update({f_scope:{}})

        if f_field not in f_lists[f_scope].keys():
            f_lists[f_scope].update({f_field:[]})

        if f_desc is True:
            f_lists[f_scope][f_field].extend(f_info["child_terms"])
        else:
            f_lists[f_scope][f_field].append(f_info["id"])

    # creating the queries & combining w/ possible existing ones
    for f_scope in f_lists.keys():
        f_s_l = []
        for f_field, f_queries in f_lists[f_scope].items():
            if len(f_queries) == 1:
                f_s_l.append({ f_field: f_queries[0] })
            else:
                f_s_l.append({ f_field: {"$in": f_queries } })

        if f_scope in byc["queries"]:
            f_s_l.append(byc["queries"][f_scope])

        if len(f_s_l) == 1:
            byc["queries"].update({ f_scope: f_s_l[0] })
        elif len(f_s_l) > 1:
            byc["queries"].update({ f_scope: { logic: f_s_l } })

    return byc

################################################################################

def update_query_for_scope( byc, query, scope, bool_mode="AND"):

    logic = boolean_to_mongo_logic(bool_mode)

    if not scope in byc["queries"]:
        byc["queries"][scope] = query
    else:
        byc["queries"][scope] = { logic: [ byc["queries"][scope], query ] }

    return byc

################################################################################

def _update_queries_from_geoquery( byc ):

    geo_q, geo_pars = geo_query( byc )

    if not geo_q:
        return byc

    update_query_for_scope( byc, geo_q, "biosamples", bool_mode="AND")

    return byc

################################################################################

def _update_queries_from_variants( byc ):

    if not "variant_request_type" in byc:
        return byc

    if not byc["variant_request_type"] in byc["variant_definitions"]["request_types"].keys():
        if not "variants" in byc["queries"]:
            return byc

    if "variantTypeRequest" in byc["variant_request_type"]:
        create_variantTypeRequest_query( byc )
    elif "variantIdRequest" in byc["variant_request_type"]:
        create_variantIdRequest_query( byc )
    elif "variantCNVrequest" in byc["variant_request_type"]:
        create_variantCNVrequest_query( byc )
    elif "variantAlleleRequest" in byc["variant_request_type"]:
        create_variantAlleleRequest_query( byc )
    elif "variantRangeRequest" in byc["variant_request_type"]:
        create_variantRangeRequest_query( byc )
    elif "geneVariantRequest" in byc["variant_request_type"]:
        create_geneVariantRequest_query( byc )
    else:
        return byc

    return byc

################################################################################

def set_pagination_range(d_count, byc):

    r_range = [
        byc["pagination"]["skip"] * byc["pagination"]["limit"],
        byc["pagination"]["skip"] * byc["pagination"]["limit"] + byc["pagination"]["limit"],
    ]

    if byc["pagination"]["skip"] == 0 and byc["pagination"]["limit"] == 0:
        byc["pagination"].update({"range":[0,d_count]})
        return byc

    r_l_i = d_count - 1

    if r_range[0] > r_l_i:
        r_range[0] = r_l_i
    if r_range[-1] > d_count:
        r_range[-1] = d_count

    byc["pagination"].update({"range":r_range})

    return byc

################################################################################

def paginate_list(this, byc):

    if byc["pagination"]["limit"] < 1:
        return this

    r = byc["pagination"]["range"]

    t_no = len(this)
    r_min = r[0] + 1
    r_max = r[-1]

    if r_min > t_no:
        return []
    if r_max > t_no:
        return this[r[0]:r_max]

    return this[r[0]:r[-1]]

################################################################################
################################################################################
################################################################################

def geo_query( byc ):

    geo_q = { }
    geo_pars = { }

    if not "geoloc_definitions" in byc:
        return geo_q, geo_pars

    g_p_defs = byc["geoloc_definitions"]["parameters"]
    g_p_rts = byc["geoloc_definitions"]["request_types"]
    geo_root = byc["geoloc_definitions"]["geo_root"]

    req_type = ""
    # TODO: Make this modular & fix the one_of interpretation to really only 1
    for rt in g_p_rts:
        g_p = { }
        min_p_no = 1
        mat_p_no = 0
        if "all_of" in g_p_rts[ rt ]:
            g_q_k = g_p_rts[ rt ]["all_of"]
            min_p_no = len(g_q_k)
        elif "one_of" in g_p_rts[ rt ]:
            g_q_k = g_p_rts[ rt ]["one_of"]
        else:
            continue

        # print(rt)
        # print(byc["form_data"]["filters"])
        all_p = g_p_rts[ rt ].get("any_of", []) + g_q_k
 
        for g_k in g_p_defs.keys():

            if g_k not in all_p:
                continue

            g_default = None
            if "default" in g_p_defs[ g_k ]:
                g_default = g_p_defs[ g_k ][ "default" ]

            # TODO: This is an ISO lower hack ...

            if g_k.lower() in byc["form_data"]:
                g_v = byc["form_data"][g_k.lower()]
            else:
                g_v = g_default
            if g_v is None:
                continue
            if not re.compile( g_p_defs[ g_k ][ "pattern" ] ).match( str( g_v ) ):
                continue
            if "float" in g_p_defs[ g_k ][ "type" ]:
                g_p[ g_k ] = float(g_v)
            else:
                g_p[ g_k ] = g_v

            if g_k in g_q_k:
                mat_p_no +=1

        if mat_p_no < min_p_no:
            continue

        req_type = rt
        geo_pars = g_p   

    if "city" in req_type:
        geo_q = return_geo_city_query(geo_root, geo_pars)
        return geo_q, geo_pars

    if "id" in req_type:
        geo_q = { "id": re.compile( geo_pars["id"], re.IGNORECASE ) }
        return geo_q, geo_pars

    if "ISO3166alpha2" in req_type:
        geo_q = { "provenance.geo_location.properties.ISO3166alpha2": byc["form_data"]["iso3166alpha2"].upper() }
        return geo_q, geo_pars

    if "geoquery" in req_type:
        geoq_l = [ return_geo_longlat_query(geo_root, geo_pars) ]
        for g_k in g_p_rts["geoquery"]["any_of"]:
            if g_k in geo_pars.keys():
                g_v = geo_pars[g_k]
                if len(geo_root) > 0:
                    geopar = ".".join( [geo_root, "properties", g_k] )
                else:
                    geopar = ".".join(["properties", g_k])
                geoq_l.append({ geopar: re.compile( r'^'+str(g_v), re.IGNORECASE ) })

        if len(geoq_l) > 1:
            geo_q = {"$and": geoq_l }
        else:
            geo_q = geoq_l[0]

    return geo_q, geo_pars

################################################################################

def return_geo_city_query(geo_root, geo_pars):

    geoq_l = []

    for g_k, g_v in geo_pars.items():

        if len(geo_root) > 0:
            geopar = ".".join( [geo_root, "properties", g_k] )
        else:
            geopar = ".".join(["properties", g_k])

        geoq_l.append( { geopar: re.compile( r'^'+str(g_v), re.IGNORECASE ) } )

    if len(geoq_l) > 1:
        return {"$and": geoq_l }
    else:
        return geoq_l[0]

################################################################################

def return_geo_longlat_query(geo_root, geo_pars):

    if len(geo_root) > 0:
        geojsonpar = ".".join( (geo_root, "geometry") )
    else:
        geojsonpar = "geo_location.geometry"

    geo_q = {
        geojsonpar: {
            '$near': SON(
                [
                    (
                        '$geometry', SON(
                            [
                                ('type', 'Point'),
                                ('coordinates', [
                                    geo_pars["geo_longitude"],
                                    geo_pars["geo_latitude"]
                                ])
                            ]
                        )
                    ),
                    ('$maxDistance', geo_pars["geo_distance"])
                ]
            )
        }
    }



    return geo_q




#!/usr/bin/env python3

import sys
from os import path, environ, pardir
from copy import deepcopy
from liftover import get_lifter
from bycon import *

# local
# pkg_path = path.join( path.dirname( path.abspath(__file__) ), pardir, pardir )
# sys.path.append( pkg_path )

"""podmd

This app is used to prototype federated Beacon queries through
translating / sending / retrieving / converting Beacon queries in v2 format to
the format of the respective Beacon instances.

Queries are expected in the GRCh38 Beacon v2 format with parameters:

* `referenceName` as either `refseq:NC_...` or just `17` etc.
* `start`
* `alternateBases`
* `referenceBases` aren't supported by most olf beacons?

TODO:

* move from the "parameters per instance" to "parameters per version"
    - this may not work if those instances take "liberties" with their parameters ...
* split into different functions
    - a main aggregator which handles i/o for a list of beacons (i.e. the original)
    - an option to have it just return the query URLs w/o executing them, e.g. as handovers
    - a single query processor, taking the modified query url as input, retrieving the
      query & providing the response in the standard v2 format
* provide a prototype front-end which parses the list of transformed queries & uses
  the retriever to asynchronously populate the front-end, and to provide an aggregate
  summary
* message if query (e.g. brackets ...) is not supported
* liftover examples ... https://pypi.org/project/liftover/ 


#### Tests

* http://progenetix.org/beacon/aggregator/?requestedGranularity=boolean&referenceBases=G&alternateBases=A&limit=1000&skip=0&datasetIds=progenetix&assemblyId=GRCh38&referenceName=refseq%3ANC_000017.11&filterLogic=AND&includeDescendantTerms=True&start=7577120

* https://cancer.sanger.ac.uk/api/ga4gh/beacon?allele=A&chrom=7&dataset=cosmic&format=json&pos=140753336&ref=38
* http://progenetix.org/services/aggregator/debug=1
* http://progenetix.org/cgi/bycon/beaconServer/aggregator.py?debug=1&reference_name=7&start=140753335&alternate_bases=A&assemblyId=GRCh38&responseEntityId=genomicVariant
* http://progenetix.org/cgi/bycon/beaconServer/aggregator.py?assemblyId=GRCh38&referenceName=17&variantType=DEL&filterLogic=AND&start=7500000&start=7676592&end=7669607&end=7800000
"""

################################################################################
################################################################################
################################################################################

def main():

    aggregator()
   
################################################################################

def beaconAggregator():
    aggregator()

################################################################################

def aggregator():

    initialize_bycon_service(byc)
    parse_filters(byc)
    parse_variant_parameters(byc)
    generate_genomic_intervals(byc)
    create_empty_service_response(byc)    

    cgi_break_on_errors(byc)

    if "selected_beacons" in byc["form_data"]:
        byc["service_config"].update({"selected_beacons": byc["form_data"]["selected_beacons"]})

    b_p = byc["service_config"]["beacon_params"]["instances"]

    byc["service_response"]["meta"]["received_request_summary"]["requested_granularity"] ="boolean"
    check_switch_to_boolean_response(byc)
    byc["service_response"].update({"response": { "response_sets": [] }})

    for b in byc["service_config"]["selected_beacons"]:

        ext_defs = b_p[b]

        for ds_id in ext_defs.get("dataset_ids", [""]):

            byc["form_data"].update({"dataset_ids": [ds_id]})

            pvs = {}

            """
            Values (for the respective parameter names) are processed step-wise:
            1. defaults are read in
            2. form values w/ remap infor are processed
            3. the non-remapped p=v are added
            4. fixed `value` parameters are (over-)written
            TODO: should we be explicit in parsing selected parameters,
            i.e. evaluating all (relevant) v2 ones?
            """

            set_dataset_id(pvs, ext_defs, ds_id)
            set_default_values(pvs, ext_defs, byc)
            remap_parameters_values(pvs, ext_defs, byc)
            add_parameter_values(pvs, ext_defs, byc)
            remap_min_max_positions(pvs, ext_defs, byc)
            set_fixed_values(pvs, ext_defs, byc)

            url = "{}{}".format(ext_defs["base_url"], urllib.parse.urlencode(pvs))
            resp_start = time.time()
            r = retrieve_beacon_response(url, byc)
            resp_end = time.time()
            # prjsoncam(r)
            r_f = format_response(r, url, ext_defs, ds_id, byc)
            r_f["info"].update({"response_time": "{}ms".format(round((resp_end - resp_start) * 1000, 0)) })
            byc["service_response"]["response"]["response_sets"].append(r_f)

    for r in byc["service_response"]["response"]["response_sets"]:
        if r["exists"] is True:
            byc["service_response"]["response_summary"].update({"exists": True})
            continue

    cgi_print_response( byc, 200 )

################################################################################
################################################################################
################################################################################

if __name__ == '__main__':
    main()



#!/usr/bin/env python3

import cgi, cgitb
import re, json
from os import path, pardir
import sys
from pymongo import MongoClient

pkg_path = path.join( path.dirname( path.abspath(__file__) ), pardir, pardir )
sys.path.append( pkg_path )
from bycon import *

"""podmd
* <https://progenetix.org/services/geolocations?city=zurich>
* <https://progenetix.org/services/geolocations?geoLongitude=8.55&geoLatitude=47.37&geoDistance=100000>
* <https://progenetix.org/services/geolocations?geoLongitude=8.55&geoLatitude=47.37&geoDistance=100000&output=map>
* <http://progenetix.org/services/geolocations?bubble_stroke_weight=0&marker_scale=5&canvas_w_px=1000&file=https://raw.githubusercontent.com/progenetix/pgxMaps/main/rsrc/locationtest.tsv&debug=&output=map&help=true>
* <http://progenetix.org/cgi/bycon/services/geolocations.py?city=New&ISO3166alpha2=UK&output=map&markerType=marker>
podmd"""

################################################################################
################################################################################
################################################################################

def main():

    geolocations()
    
################################################################################

def geolocations():

    initialize_bycon_service(byc)

    # for the geolocs database, not the provenance object
    byc["geoloc_definitions"]["geo_root"] = "geo_location"
    
    create_empty_service_response(byc)

    # TODO: move the map table reading to a sane place 
    if "file" in byc["form_data"]:
        results = read_geomarker_table_web(byc)

    else:

        query, geo_pars = geo_query( byc )
        for g_k, g_v in geo_pars.items():
            response_add_received_request_summary_parameter(byc, g_k, g_v)

        if len(query.keys()) < 1:
            response_add_error(byc, 422, "No query generated - missing or malformed parameters" )
        
        cgi_break_on_errors(byc)

        results, e = mongo_result_list( byc["dataset_ids"][0], byc["geo_coll"], query, { '_id': False } )
        response_add_error(byc, 422, e)

    print_map_from_geolocations(byc, results)

    if len(results) == 1:
        if "geo_distance" in byc["form_data"]:
            l_l = results[0]["geo_location"]["geometry"]["coordinates"]
            geo_pars = {
                "geo_longitude": l_l[0],
                "geo_latitude": l_l[1],
                "geo_distance": int(byc["form_data"]["geo_distance"])
            }
            query = return_geo_longlat_query(byc["geoloc_definitions"]["geo_root"], geo_pars)
            results, e = mongo_result_list( byc["dataset_ids"][0], byc["geo_coll"], query, { '_id': False } )
            response_add_error(byc, 422, e)
    
    cgi_break_on_errors(byc)

    if "text" in byc["output"]:
        open_text_streaming(byc["env"], "browser")
        for g in results:
            s_comps = []
            for k in ["city", "country", "continent"]:
                s_comps.append(str(g["geo_location"]["properties"].get(k, "")))
            s_comps.append(str(g.get("id", "")))
            for l in g["geo_location"]["geometry"].get("coordinates", [0,0]):
                s_comps.append(str(l))
            print("\t".join(s_comps))
        exit()

    populate_service_response( byc, results)
    cgi_print_response( byc, 200 )

################################################################################
################################################################################

if __name__ == '__main__':
    main()



from pymongo import MongoClient
from bson.son import SON
from uuid import uuid4
# import logging
import datetime
import sys

from cgi_parsing import cgi_debug_message, test_truthy

################################################################################

def mongo_result_list(db_name, coll_name, query, fields):

    results = [ ]
    error = False

    mongo_client = MongoClient( )

    try:
        results = list( mongo_client[ db_name ][ coll_name ].find( query, fields ) )
    except Exception as e:
        error = e

    mongo_client.close( )
 
    return results, error

################################################################################

def process_empty_request(ds_id, collname, byc, ret_no=10):

    mongo_client = MongoClient()
    data_db = mongo_client[ ds_id ]
    data_coll = mongo_client[ ds_id ][ collname ]

    c = data_coll.estimated_document_count()
    r = list( data_coll.aggregate([{"$sample": {"size":ret_no}}]) )

    return r, c

################################################################################

def execute_bycon_queries(ds_id, byc):

    max_bs_number_for_v_in_query = 2500

    # last_time = datetime.datetime.now()
    # logging.info("\t start query: {}".format(last_time))

    """podmd
    
    Pre-configured queries are performed in an aggregation pipeline against
    the standard "Progenetix"-type MongoDB collections.
        
    podmd"""

    h_o_defs = byc[ "handover_definitions" ]["h->o_methods"]

    exe_queries = { }
    if not "dataset_results" in byc.keys():
        byc.update({"dataset_results":{}})

    data_client = MongoClient( )
    data_db = data_client[ ds_id ]
    data_collnames = data_db.list_collection_names()

    ho_client = MongoClient()
    ho_db = ho_client[ byc["config"]["info_db"] ]
    ho_collname = byc["config"][ "handover_coll" ]
    ho_coll = ho_db[ ho_collname ]

    if test_truthy( byc["form_data"].get("annotated_only", False) ):
        if "variants" in byc["queries"]:
            byc["queries"].update({"variants": { "$and": [ byc["queries"]["variants"], {"info.annotation_derived":True} ] } } )
        else:
            byc["queries"].update( {"variants": {"info.annotation_derived":True} } )


    for collname in byc[ "queries" ].keys():
        if collname in byc[ "config" ][ "queried_collections" ]:
            exe_queries[ collname ] = byc[ "queries" ][ collname ]

    byc.update({"queries_at_execution": exe_queries})

    if byc["original_queries"] is None:
        byc.update({"original_queries": exe_queries})

    # collection of results
    prefetch = { }
    prevars = { 
        "ds_id": ds_id,
        "data_db": data_db,
        "h_o_defs": h_o_defs,
        "original_queries": byc["original_queries"],
        "pref_m": "", "query": { }
    }

    ############################################################################

    cgi_debug_message(byc, "queries at execution", exe_queries)

    ############################################################################

    if "variant_annotations" in exe_queries:

        prevars["pref_m"] = "variant_annotations._id"
        prevars["query"] = exe_queries[ "variant_annotations" ]
        prefetch.update( { prevars["pref_m"]: _prefetch_data(prevars) } )
        byc["dataset_results"].update( { ds_id: prefetch } )

        return byc

    ############################################################################

    """podmd
    
    All queries are aggregated towards biosamples.
        
    podmd"""

    if "biosamples" in exe_queries:

        prevars["pref_m"] = "biosamples.id"
        prevars["query"] = exe_queries[ "biosamples" ]
        prefetch.update( { prevars["pref_m"]: _prefetch_data(prevars) } )

    if "individuals" in exe_queries:

        prevars["pref_m"] = "individuals.id"
        prevars["query"] = exe_queries[ "individuals" ]
        prefetch.update( { prevars["pref_m"]: _prefetch_data(prevars) } )

        prevars["pref_m"] = "biosamples.id"
        prevars["query"] = { "individual_id": { '$in': prefetch["individuals.id"]["target_values"] }  }
        biosids_from_indq =  _prefetch_data(prevars)

        if "biosamples.id" in prefetch:
            bsids = list( set( biosids_from_indq["target_values"] ) & set( prefetch["biosamples.id"]["target_values"] ) )
            prefetch[ "biosamples.id" ].update( { "target_values": bsids, "target_count": len(bsids) } )
        else:
            prefetch["biosamples.id"] = biosids_from_indq

    if "callsets" in exe_queries:

        # since callsets contain biosample_id no double calling is required
        prevars["pref_m"] = "callsets.biosample_id->biosamples.id"
        prevars["query"] = exe_queries[ "callsets" ]
        prefetch.update( { prevars["pref_m"]: _prefetch_data(prevars) } )

        if "biosamples.id" in prefetch:
            bsids = list( set( prefetch["callsets.biosample_id->biosamples.id"]["target_values"] ) & set( prefetch["biosamples.id"]["target_values"] ) )
            prefetch[ "biosamples.id" ].update( { "target_values": bsids, "target_count": len(bsids) } )
        else:
            prefetch["biosamples.id"] = prefetch["callsets.biosample_id->biosamples.id"]

    if "variants" in exe_queries:

        if exe_queries["variants"]:

            """podmd
            ### `variants` Query and Aggregation

            1. If a `variants` query exists (i.e. has been defined in `exe_queries`), in a first pass
            all `biosample_id` values are retrieved.
            2. If already a `"biosamples.id"` result exists (e.g. from a biosample query), the lists
            of callset `id` values from the different queries are intersected. Otherwise, the callsets
            from the variants query are the final ones.
            3. Since so far not all matching variants have been retrieved (only the biosamples which
            contain them), they are now fetched using the original query or a combination of the
            original query and the matching biosamples from the intersect.
            podmd"""

            prevars["pref_m"] = "variants.biosample_id->biosamples.id"
            prevars["query"] = exe_queries[ "variants" ]
            prefetch.update( { prevars["pref_m"]: _prefetch_data(prevars) } )

            if "biosamples.id" in prefetch:
                bsids = list( set( prefetch["biosamples.id"]["target_values"] ) & set( prefetch["variants.biosample_id->biosamples.id"]["target_values"] ) )
                prefetch[ "biosamples.id" ].update( { "target_values": bsids, "target_count": len(bsids) } )
                exe_queries[ "variants" ] = { "$and": [ exe_queries[ "variants" ], { "biosample_id": { "$in": bsids } } ] }
            else:
                prefetch[ "biosamples.id" ] = prefetch["variants.biosample_id->biosamples.id"]

            prevars["pref_m"] = "variants._id"
            prevars["query"] = exe_queries[ "variants" ]
            prefetch.update( { prevars["pref_m"]: _prefetch_data(prevars) } )

            prevars["pref_m"] = "variants.variant_internal_id"
            prevars["query"] = { "_id": { "$in": prefetch[ "variants._id" ]["target_values"] } }
            prefetch.update( { prevars["pref_m"]: _prefetch_data(prevars) } )


    ############################################################################
    """podmd
    ### Result Aggregation

    The above queries have provided `biosamples.id` values which now are used to retrieve the
    matching final biosample and callset `_id` values.

    For variants the `_id` values only exist if a variants query had been performed.
    In that case no separate recall has to be performed since a biosample intersection
    had been performed already (to limit the a priori variant response).

    If no variant query was performed _but_ the response asks for variants => all
    callset variants will be returned.

    TODO: Benchmark if the `_id` retrieval & storage speeds up biosample and callset recovery
    in handover scenarios or if `id` is fine.

    TODO: The return-driven query selection will need to be refined; queries can
    potentially lead to huge responses here...
    podmd"""

    if not "biosamples.id" in prefetch:

        prefetch.update({"biosamples.id": { **prevars["h_o_defs"]["biosamples.id"]}})
        prefetch["biosamples.id"].update({
            "id": str(uuid4()),
            "source_db": ds_id,
            "target_values": [],
            "target_count": 0
            })

        byc["dataset_results"].update( { ds_id: prefetch } )
        return byc

    prevars["pref_m"] = "callsets._id"
    prevars["query"] = { "biosample_id": { "$in": prefetch[ "biosamples.id" ]["target_values"] } }
    prefetch.update( { prevars["pref_m"]: _prefetch_data(prevars) } )

    prevars["pref_m"] = "biosamples._id"
    prevars["query"] = { "id": { "$in": prefetch[ "biosamples.id" ]["target_values"] } }
    prefetch.update( { prevars["pref_m"]: _prefetch_data(prevars) } )

    # TODO: have this checked... somewhere else based on the response_entity_id
    if "response_entity_id" in byc:
        if "individual" in byc["response_entity_id"] or "phenopacket" in byc["response_entity_id"]:
            _prefetch_add_individuals(prevars, prefetch)
        elif "genomicVariant" in byc["response_entity_id"] and not "variants._id"  in prefetch:
            _prefetch_add_all_sample_variants(prevars, prefetch)

    if "variant_annotations" in data_collnames:
        if "variants._id" in prefetch:
            _prefetch_add_variant_annotations(prevars, prefetch)
    
    ############################################################################

    data_client.close( )
    ho_client.close( )

    byc["dataset_results"].update( { ds_id: prefetch } )

    return byc

################################################################################

def _prefetch_data(prevars):

    """podmd
    The prefetch pref_m queries the specified collection `source_collection` of
    the `data_db` with the provided query, and stores the distinct values of the
    `source_key` as `target_values`.

    The results may reference across collections. A typical example here would be
    to retrieve `biosample_id` values from the `variants` collection to point
    to `id` values in the `biosamples` collection.

    These "handover" objects can then be stored and used to retrieve values of
    previous queries for procedural use or second-pass data retrieval.

    podmd"""

    pref_m = prevars["pref_m"]
    data_db = prevars["data_db"]
    h_o_defs = prevars["h_o_defs"][pref_m]

    # print(h_o_defs["source_collection"])
    # print(h_o_defs["source_key"])
    # print(prevars["query"])

    dist = data_db[ h_o_defs["source_collection"] ].distinct( h_o_defs["source_key"], prevars["query"] )

    h_o = { **h_o_defs }
    h_o.update(
        {
            "id": str(uuid4()),
            "source_db": prevars["ds_id"],
            "target_values": dist,
            "target_count": len(dist),
            "original_queries": prevars.get("original_queries", None)
        }
    )

    return h_o

################################################################################

def _prefetch_vars_from_biosample_loop( prevars ):

    # TODO: This still allows unlimited variants, e.g. all from thousands
    # of samples and will let the server time out if too many...
    # A new paradigm is needed which includes the pagination at this step.

    pref_m = prevars["pref_m"]
    data_db = prevars["data_db"]
    h_o_defs = prevars["h_o_defs"][pref_m]

    h_o = { **h_o_defs }
    h_o["target_values"] = [ ]

    for bs_id in prevars["query"]["biosample_id"]["$in"]:
        for v in data_db[ "variants" ].find( { "biosample_id": bs_id} ):
            h_o["target_values"].append( v["_id"])

    h_o.update(
        {
            "id": str(uuid4()),
            "source_db": prevars["ds_id"],
            "target_count": len(h_o["target_values"]),
            "original_queries": prevars.get("original_queries", None)
        }
    )

    return h_o

################################################################################

def _prefetch_add_individuals(prevars, prefetch):

    prevars["pref_m"] = "biosamples.individual_id->individuals.id"
    prevars["query"] = { "_id": { "$in": prefetch[ "biosamples._id" ]["target_values"] } }
    prefetch.update( { "individuals.id": _prefetch_data(prevars) } )

    prevars["pref_m"] = "individuals._id"
    prevars["query"] = { "id": { "$in": prefetch[ "individuals.id" ]["target_values"] } }
    prefetch.update( { prevars["pref_m"]: _prefetch_data(prevars) } )

    return prefetch

################################################################################

def _prefetch_add_all_sample_variants(prevars, prefetch):

    prevars["pref_m"] = "variants._id"
    prevars["query"] = { "biosample_id": { "$in": prefetch[ "biosamples.id" ]["target_values"] } }
    prefetch.update( { prevars["pref_m"]: _prefetch_vars_from_biosample_loop( prevars ) } )
    # print("Storage size for {} entries in {}: {}".format(prefetch[prevars["pref_m"]]["target_count"], prevars["pref_m"], sys.getsizeof(prefetch[prevars["pref_m"]]["target_values"]) / 1000000))

    return prefetch

################################################################################

def _prefetch_add_variant_annotations(prevars, prefetch):

    prevars["pref_m"] = "variants.variantannotation_id->variant_annotations.id"
    prevars["query"] = { "_id": { "$in": prefetch[ "variants._id" ]["target_values"] } }
    prefetch.update( { prevars["pref_m"]: _prefetch_data(prevars) } )

    prevars["pref_m"] = "variant_annotations._id"
    prevars["query"] = { "id": { "$in": prefetch[ "variants.variantannotation_id->variant_annotations.id" ]["target_values"] } }
    prefetch.update( { prevars["pref_m"]: _prefetch_data(prevars) } )

    return prefetch





#!/usr/bin/env python3

import sys
from os import path, environ, pardir
from copy import deepcopy
from liftover import get_lifter

pkg_path = path.join( path.dirname( path.abspath(__file__) ), pardir, pardir )
sys.path.append( pkg_path )
from bycon import *

"""podmd

#### Tests

* http://progenetix.org/cgi/bycon/beaconServer/retriever.py?debug=&selectedBeacons=progenetixTest&url=http%3A//progenetix.test/beacon/g_variants/%3FdatasetIds%3Dprogenetix%26assemblyId%3DGRCh38%26referenceName%3D17%26variantType%3DDEL%26start%3D7500000%252C7676592%26end%3D7669607%252C7800000
"""

################################################################################
################################################################################
################################################################################

def main():

    retriever()
   
################################################################################

def retriever():

    initialize_bycon_service(byc, "aggregator")
    parse_filters(byc)
    parse_variant_parameters(byc)
    generate_genomic_intervals(byc)
    create_empty_service_response(byc)    

    cgi_break_on_errors(byc)

    b = byc["form_data"].get("selected_beacons", [])
    url = byc["form_data"].get("url", "")
    # print(url)
    if len(b) != 1:
        print_text_response('not a single "selectedBeacons" value')
    byc["service_config"].update({"selected_beacons": b})
    b = b[0]
    if not "http" in url:
        print_text_response('url seems missing / incomplete')    
    b_p = byc["service_config"]["beacon_params"]["instances"]
    if not b in b_p.keys():
        print_text_response(f'"{b}"is not in available beacon definitions')

    byc["service_response"]["meta"]["received_request_summary"]["requested_granularity"] ="boolean"
    check_switch_to_boolean_response(byc)
    byc["service_response"].update({"response": { "response_sets": [] }})

    ext_defs = b_p[b]

    # TODO: extract dataset id from URL using the ext_defs parameter mapping
    ds_id = byc["form_data"].get("dataset_ids", [])
    if len(ds_id) != 1:
        ds_id = ext_defs["dataset_ids"]
    ds_id = ds_id[0]

    resp_start = time.time()
    # print_text_response(url)
    r = retrieve_beacon_response(url, byc)
    resp_end = time.time()
    # prjsoncam(r)
    # print(url)
    r_f = format_response(r, url, ext_defs, ds_id, byc)
    r_f["info"].update({"response_time": "{}ms".format(round((resp_end - resp_start) * 1000, 0)) })
    byc["service_response"]["response"]["response_sets"].append(r_f)

    for r in byc["service_response"]["response"]["response_sets"]:
        if r["exists"] is True:
            byc["service_response"]["response_summary"].update({"exists": True})
            continue

    cgi_print_response( byc, 200 )

################################################################################
################################################################################
################################################################################

if __name__ == '__main__':
    main()



import cgi, cgitb
import re, yaml
from pymongo import MongoClient

from cgi_parsing import *

################################################################################

def parse_filters(byc):

    if "filters" in byc["form_data"]:
        fs = byc["form_data"]["filters"]
        fs = check_filter_values(fs, byc)
        if len(fs) > 0:
            byc.update( { "filters": fs } )
            return byc
            
    return byc

################################################################################

def get_filter_flags(byc):

    ff = {
        "logic": byc[ "config" ][ "filter_flags" ][ "logic" ],
        "precision": byc[ "config" ][ "filter_flags" ][ "precision" ],
        "descendants": byc[ "config" ][ "filter_flags" ][ "include_descendant_terms" ]
    }

    if "form_data" in byc:
        if "filter_logic" in byc[ "form_data" ]:
            ff["logic"] = boolean_to_mongo_logic( byc["form_data"]['filter_logic'] )
        if "filter_precision" in byc[ "form_data" ]:
            ff["precision"] = byc["form_data"]['filter_precision']
        if "include_descendant_terms" in byc[ "form_data" ]:
            i_d_t = str(byc[ "form_data" ].get("include_descendant_terms", 1)).lower()
            if i_d_t in ["0", "-1", "no", "false"]:
                ff["descendants"] = False

    byc.update( { "filter_flags": ff } )

    return byc

################################################################################

def check_filter_values(filters, byc):

    f_defs = byc["filter_definitions"]

    checked = [ ]
    for f in filters:
        if not isinstance(f, dict):
            f = {"id":f}
        if not "id" in f:
            continue
        pre = re.split('-|:', f["id"])[0]
        for f_t, f_d in f_defs.items():
            if re.compile( f_d["pattern"] ).match( f["id"] ):       
                if f not in checked:
                    checked.append( f )

    return checked
  
################################################################################

def select_dataset_ids(byc):

    if not "dataset_ids" in byc.keys():
        byc.update( { "dataset_ids": [ ] } )

    if ds_id_from_rest_path_value(byc) is not False:
        return byc

    if ds_id_from_accessid(byc) is not False:
        return byc            

    if ds_id_from_form(byc) is not False:
        return byc            
    
    return byc

################################################################################

def ds_id_from_rest_path_value(byc):

    ds_id = rest_path_value("datasets")
    if ds_id is "empty_value":
        return False

    if ds_id not in byc["dataset_definitions"].keys():
        return False

    byc.update( { "dataset_ids": [ ds_id ] } )
    return byc

################################################################################

def ds_id_from_accessid(byc):

    # TODO: This is very verbose. In principle there should be an earlier
    # test of existence...

    accessid = byc["form_data"].get("accessid", None)
    if "accessid" is None:
        return False

    info_db = byc["config"].get("info_db", None)
    if "info_db" is None:
        return False

    ho_collname = byc["config"].get("handover_coll", None)
    if "ho_collname" is None:
        return False

    ho_client = MongoClient()
    h_o = ho_client[ info_db ][ ho_collname ].find_one( { "id": accessid } )
    if not h_o:
        return False

    ds_id = h_o.get("source_db", None)
    if "ds_id" is None:
        return False

    if ds_id not in byc["dataset_definitions"].keys():
        return False

    byc.update( { "dataset_ids": [ ds_id ] } )
    return byc

################################################################################

def ds_id_from_form(byc):

    f_ds_ids = byc["form_data"].get("dataset_ids", None)
    if f_ds_ids is None:
        return False

    ds_ids = [ ]
    for ds_id in f_ds_ids:
        if ds_id in byc["dataset_definitions"].keys():
            ds_ids.append(ds_id)

    if len(ds_ids) < 1:
        return False

    byc.update( { "dataset_ids": ds_ids } )
    
    return byc

################################################################################



import cgi, cgitb, json, re, sys
from urllib.parse import urlparse, parse_qs, unquote
from os import environ
from humps import camelize, decamelize

################################################################################

def set_debug_state(debug=0):

    if test_truthy(debug):
        cgitb.enable()
        print('Content-Type: text')
        print()
        return True

    elif environ.get('REQUEST_URI'):
        if "debug=1" in environ.get('REQUEST_URI'):
            cgitb.enable()
            print('Content-Type: text')
            print()
            return True

    return False

################################################################################

def boolean_to_mongo_logic( logic="AND" ):

    if "OR" in logic.upper():
        return '$or'
    return '$and'

################################################################################

def select_this_server( byc ):

    s_uri = str(environ.get('SCRIPT_URI'))
    # print(s_uri)

    # TODO

    if "progenetix.test" in s_uri:
        if "https:" in s_uri:
            return "https://progenetix.test"
        else:
            return "http://progenetix.test" 
    elif "cancercelllines.test" in s_uri:
        if "https:" in s_uri:
            return "https://cancercelllines.test"
        else:
            return "http://cancercelllines.test" 

    # TODO: ERROR hack for https/http mix, CORS...
    # ... since cloudflare provides https mapping
    # if "https:" in s_uri:
    if "http:" in s_uri:
        return "https://"+str(environ.get('HTTP_HOST'))
    else:
        return "http://"+str(environ.get('HTTP_HOST'))

################################################################################

def cgi_parse_query(byc):
    
    r_m = environ.get('REQUEST_METHOD', '')
    if "POST" in r_m:
        cgi_parse_POST(byc)
    else:
        cgi_parse_GET(byc)

################################################################################

def cgi_parse_POST(byc):

    content_len = environ.get('CONTENT_LENGTH', '0')
    content_typ = environ.get('CONTENT_TYPE', '')

    defs = byc.get("beacon_defaults", {})
    form = {}

    # TODO: catch error & return for non-json posts
    if "json" in content_typ:
        body = sys.stdin.read(int(content_len))
        jbod = json.loads(body)
        if "debug" in jbod:
            if jbod["debug"] > 0:                 
                byc.update({"debug_mode": set_debug_state(1)})

        # TODO: this hacks the v2 structure; ideally should use requestParameters schemas
        if "query" in jbod:
            for p, v in jbod["query"].items():
                if p == "requestParameters":
                    for rp, rv in v.items():
                        rp_d = decamelize(rp)
                        if "datasets" in rp:
                            if "datasetIds" in rv:
                                form.update({"dataset_ids": rv["datasetIds"]})
                        elif "g_variant" in rp:
                            for vp, vv in v[rp].items():
                                vp_d = decamelize(vp)
                                form.update({vp_d: vv})
                        else:
                            form.update({rp_d: rv})
                else:
                    p_d = decamelize(p)
                    form.update({p_d: v})

        # TODO: define somewhere else with proper defaults
        form.update({
            "requested_granularity": jbod.get("requestedGranularity", defs.get("requested_granularity", "record")),
            "include_resultset_responses": jbod.get("includeResultsetResponses", defs.get("include_resultset_responses", "HIT")),
            "include_handovers": jbod.get("includeHandovers", defs.get("include_handovers", False)),
            "filters": jbod.get("filters", [] )
        })
        # print(form)
        # exit()

        # transferring pagination where existing to standard form values
        pagination = jbod.get("pagination", {})
        for p_k in ["skip", "limit"]:
            if p_k in pagination:
                if re.match(r'^\d+$', str(pagination[p_k])):
                    form.update({p_k: pagination[p_k]})
        byc.update({
            "form_data": form,
            "query_meta": jbod.get("meta", {})
        })

    return byc

################################################################################

def cgi_parse_GET(byc):

    defs = byc.get("beacon_defaults", {})
    form = {}

    byc.update({"debug_mode": set_debug_state()})
    get = cgi.FieldStorage()

    for p in get:
        p_d = decamelize(p)
        if p in byc["config"]["form_list_pars"]["items"]:
            form.update({p_d: form_return_listvalue( get, p )})
        else:
            v = get.getvalue(p)
            # making sure double entries are forced to single
            if type(v) is list:
                form.update({p_d: v[0]})
            else:
                form.update({p_d: v})

    #TODO: re-evaluate hack of empty filters which avoids dirty errors downstream
    if not "filters" in form:
        form.update({"filters": []})

    form.update({
        "requested_granularity": get.getvalue("requestedGranularity", defs.get("requested_granularity", "record")),
        "include_resultset_responses": get.getvalue("includeResultsetResponses", defs.get("include_resultset_responses", "HIT")),
        "include_handovers": get.getvalue("includeHandovers", defs.get("include_handovers", False))
    })

    if "requested_schema" in form:
        try:
            byc["query_meta"].update({
                "requested_schemas": [ {"entity_type": form["requested_schema"] } ]
            } )
        except:
            pass

    byc.update({ "form_data": form })
    
    return byc

################################################################################

def rest_path_elements(byc):

    r_p_r = byc.get("request_path_root", "beacon")

    if not environ.get('REQUEST_URI'):
        return byc

    url_comps = urlparse( environ.get('REQUEST_URI') )
    url_p = url_comps.path
    p_items = re.split('/', url_p)

    if not r_p_r in p_items:
        return byc
    
    for d_k in ["debug=1", "debug=true"]:
        if d_k in p_items:
            p_items.remove(d_k)

    p_items = list(filter(None, p_items))

    r_i = p_items.index(r_p_r)

    for p_k in ["request_entity_path_id", "request_entity_path_id_value", "response_entity_path_id"]:

        r_i += 1
        if r_i >= len(p_items):
            return byc
        byc.update({p_k:p_items[r_i]})

    return byc

################################################################################

def rest_path_value(key=""):

    r_p_v = "empty_value"

    if not environ.get('REQUEST_URI'):
        return r_p_v

    url_comps = urlparse( environ.get('REQUEST_URI') )
    url_p = url_comps.path
    p_items = re.split('/', url_p)

    if "debug=1" in p_items:
        p_items.remove("debug=1")

    i = 0
    f = ""

    if len(p_items[-1]) < 2:
        del p_items[-1]

    for p in p_items:

        i += 1
        if len(p_items) > i:
            if unquote(p) in [key, key+".py", unquote(key)]:
                return unquote(p_items[ i ])
        elif p == key:
            return r_p_v

    return r_p_v

################################################################################

def form_return_listvalue( form_data, parameter ):

    l_v = [ ]
    if len(form_data) > 0:
        if parameter in form_data:
            v = form_data.getlist( parameter )
            if "null" in v:
                v.remove("null")
            if "undefined" in v:
                v.remove("undefined")
            if len(v) > 0:
                l_v  = ','.join(v)
                l_v  = l_v.split(',')

    return l_v

################################################################################

def test_truthy(this):

    if str(this).lower() in ["1", "true", "y", "yes"]:
        return True    

    return False

################################################################################

def cgi_simplify_response(byc):

    r = byc["service_response"].get("response", "ERROR: No response element in error_response")

    if "result_sets" in r:
        r_s = r["result_sets"][0]
        byc.update({ "service_response": r_s.get("results", [])})
    else:
        byc.update({ "service_response": r })

    return byc

################################################################################

def cgi_break_on_errors(byc):

    e_c = byc["error_response"]["error"].get("error_code", 200)

    # TODO: temp hack
    for k in byc["service_response"].keys():
        if "any_of" in byc["service_response"][k]:
            byc["service_response"][k].pop("any_of")
        if "all_of" in byc["service_response"][k]:
            byc["service_response"][k].pop("all_of")

    if e_c > 200:
        cgi_print_response( byc, e_c )

################################################################################

def cgi_debug_message(byc, label, debug_object):

    try:
        if byc["debug_mode"]:
            print("{}:\n\n{}\n\n".format(label, debug_object))
    except:
        pass

################################################################################

def switch_to_wrong_service_response(byc):

    byc.update( {
        "service_response": {
            "response" : {
                "error" : {
                    "error_code": 422,
                    "error_message": "No correct service path provided. Please refer to the documentation at http://info.progenetix.org/tags/Beacon"
                    },
                }
            }
        })

    return byc

################################################################################

def cgi_print_response(byc, status_code):

    r_f = ""
    f_d = {}

    delint_response(byc)

    if "form_data" in byc:
        f_d = byc["form_data"]

    if "responseFormat" in f_d:
        r_f = f_d["responseFormat"]

    # This is a simple "de-jsonify", intended to be used for already
    # pre-formatted list-like items (i.e. lists only containing objects)
    # with simple key-value pairs)
    # TODO: universal text table converter ... partially implemented

    if "text" in byc["output"]:

        cgi_simplify_response(byc)

        if isinstance(byc["service_response"], dict):
            resp = json.dumps(camelize(byc["service_response"]["response"]), default=str)
        else:
            resp = byc["service_response"]
        if isinstance(resp, list):
            l_d = [ ]
            for dp in resp:
                v_l = [ ]
                for v in dp.values():
                    print(v)
                    v_l.append(str(v))
                l_d.append("\t".join(v_l))
            resp = "\n".join(l_d)

        print_text_response(resp, byc["env"], status_code)

    if "handoversonly" in byc["output"]:
        try:        
            if "result_sets" in byc["service_response"]["response"]:
                for rs_i, rs in enumerate(byc["service_response"]["response"]["result_sets"]):
                    byc["service_response"]["response"]["result_sets"][rs_i].update({"results":[]})
        except:
            pass

    if "simple" in r_f:
        cgi_simplify_response(byc)

    response_clean_legacy(byc)
    update_error_code_from_response_summary(byc)
    switch_to_error_response(byc)
    print_json_response(byc["service_response"], byc["env"])

################################################################################

def response_clean_legacy(byc):

    legacy = ["result_sets", "data"]

    for k in legacy:
        byc["service_response"].pop(k, None)

    return byc

################################################################################

def update_error_code_from_response_summary(byc):

    if not "response_summary" in byc["service_response"]:
        return byc

    if not "exists" in byc["service_response"]["response_summary"]:
        return byc

################################################################################

def switch_to_error_response(byc):

    e_c = byc["error_response"]["error"].get("error_code", 200)

    if e_c > 200:
        if "meta" in byc["service_response"]:
            byc["error_response"].update({ "meta": byc["service_response"]["meta"]})
        byc["service_response"] = byc["error_response"]

    return byc

################################################################################

def check_switch_to_boolean_response(byc):

    try:
        if "boolean" in byc["service_response"]["meta"]["received_request_summary"].get("requested_granularity", "record"):
            byc["service_response"].pop("response", None)
            byc["service_response"]["response_summary"].pop("num_total_results", None)
            byc["service_response"]["meta"].update({"returned_granularity": "boolean"})
    except:
        pass

    return byc

################################################################################

def check_switch_to_count_response(byc):

    try:
        if "count" in byc["service_response"]["meta"]["received_request_summary"].get("requested_granularity", "record"):
            byc["service_response"].pop("response", None)
            byc["service_response"]["meta"].update({"returned_granularity": "count"})
    except:
        pass

    return byc

################################################################################

def delint_response(byc):

    try:
        if len(byc["service_response"]["beacon_handovers"]) < 1:
            byc["service_response"].pop("beacon_handovers", None)
        if len(byc["service_response"]["info"].keys()) < 1:
            byc["service_response"].pop("info", None)
    except:
        pass

    return byc

################################################################################
################################################################################
################################################################################

def open_json_streaming(byc, filename="data.json"):

    meta = byc["service_response"].get("meta", {})

    if not "local" in byc["env"]:
        print('Content-Type: application/json')
        print('Content-Disposition: attachment; filename="{}"'.format(filename))
        print('status: 200')
        print()

    print('{"meta":', end = '')
    print(json.dumps(camelize(meta), indent=None, sort_keys=True, default=str), end=",")
    print('"response":{', end='')
    for r_k, r_v in byc["service_response"].items():
        if "results" in r_k:
            continue
        if "meta" in r_k:
            continue
        print('"'+r_k+'":', end='')
        print(json.dumps(camelize(r_v), indent=None, sort_keys=True, default=str), end=",")
    print('"results":[', end="")

################################################################################

def close_json_streaming():

    print(']}}')
    exit()

################################################################################

def open_text_streaming(env="server", filename="data.pgxseg"):

    if not "local" in env:
        print('Content-Type: text/plain')
        if not "browser" in filename:
            print('Content-Disposition: attachment; filename="{}"'.format(filename))
        print('status: 200')
        print()

################################################################################

def close_text_streaming():

    print()
    exit()

################################################################################

def prjsoncam(this):
    prjsonnice(camelize(this))

################################################################################

def prjsonnice(this):
    print(decamelize_words(json.dumps(this, indent=4, sort_keys=True, default=str))+"\n")

################################################################################

def decamelize_words(j_d):

    de_cams = ["gVariants", "sequenceId", "relativeCopyClass", "speciesId", "chromosomeLocation", "genomicLocation"]
    for d in de_cams:
        j_d = re.sub(r"\b{}\b".format(d), decamelize(d), j_d)

    return j_d

################################################################################

def print_json_response(this={}, env="server", status_code=200):

    if not "local" in env:
        print('Content-Type: application/json')
        print('status:'+str(status_code))
        print()
    
    prjsoncam(this)
    print()
    exit()

######################################logi  ##########################################

def print_text_response(this="", env="server", status_code=200):

    if not "local" in env:
        print('Content-Type: text/plain')
        print('status:'+str(status_code))
        print()
    
    print(this)
    print()
    exit()

################################################################################

def print_uri_rewrite_response(uri_base="", uri_stuff="", output_par="empty_value"):

    print("Status: 302")
    print("Location: {}{}".format(uri_base, uri_stuff))
    print()
    exit()






import re, json, yaml
from os import path, scandir, pardir
from json_ref_dict import RefDict, materialize
from humps import decamelize

from cgi_parsing import prjsonnice

# local
lib_path = path.dirname( path.abspath(__file__) )
pkg_path = path.join( lib_path, pardir )

################################################################################

def read_schema_file(schema_name, item, byc, ext="json"):

    b_p_m = byc["beacon_mappings"]["default_schema_from_model"]

    schema_name = b_p_m.get(schema_name, schema_name)
    
    s_f_p = get_schema_file_path(schema_name, byc, ext)

    if byc["debug_mode"] is True:
        print(schema_name, s_f_p)

    if not s_f_p is False:
        if len(item) > 1:
            s_f_p = s_f_p+"#/"+item
        root_def = RefDict(s_f_p)
        
        exclude_keys = [ "examples" ] #"format",
        s = materialize(root_def, exclude_keys=exclude_keys)
        assert isinstance(s, dict)
        return s

    return False

################################################################################

def get_schema_file_path(schema_name, byc, ext="json"):

    config = byc["config"]

    for s_p in config["schema_paths"]["items"]:

        p = path.join( pkg_path, *config["schemas_root"], *s_p )

        s_ds = [ d.name for d in scandir(p) if d.is_dir() ]
        if schema_name in s_ds:
            s_f_p = path.join( p, schema_name, "defaultSchema."+ext )
            return s_f_p

        s_fs = [ f.name for f in scandir(p) if f.is_file() ]
        s_fs = [ f for f in s_fs if f.endswith( ext ) ]
        s_fs = [ f for f in s_fs if not f.startswith("_") ]

        for s_f in s_fs:

            f_name = path.splitext( s_f )[0]
            # print(schema_name, f_name)

            if f_name == schema_name:
                s_f_p = path.join( p, s_f )
                return s_f_p

    return False

################################################################################

def instantiate_schema(schema, is_root=True):

    if is_root is not True:
        if 'type' in schema.keys():

            t = schema['type']
        
            if t == 'array' or t == 'list':
                schema = []
            elif t == 'object':
                schema = { }
            elif t == 'integer':
                schema = int()
            elif t == 'number':
                schema = float()
            elif t == 'boolean':
                schema = False
            else:
                schema = ""
               
            return schema
      
    else:
        for k, val in schema.items():
            if isinstance(val, dict):
                schema.update({ k: instantiate_schema(val, False) })
                
    return schema
        
################################################################################

def create_empty_instance(schema):

    s_i = instantiate_schema(schema, True)
    s_i = decamelize(s_i)
    return s_i

################################################################################

def object_instance_from_schema_name(byc, schema_name, root_key, ext="json"):

    s_f = read_schema_file(schema_name, root_key, byc, ext)
    s_i = create_empty_instance( s_f )
    return s_i

################################################################################
